package auth

import (
	"crypto/md5"
	"encoding/hex"
	"fmt"
	"goapp/config"
	"goapp/database"
	"goapp/models"
	"os"
	"strings"
	"time"

	"github.com/dgrijalva/jwt-go"
	"github.com/gofiber/fiber/v2"
	"github.com/google/uuid"
)

type TokenDetails struct {
	AccessToken  string
	RefreshToken string
	AccessUuid   string
	RefreshUuid  string
	AtExpires    int64
	RtExpires    int64
}

type Tokens struct {
	AccessToken  string `json:"access_token"`
	RefreshToken string `json:"refresh_token"`
}

type AuthLogin struct {
	Login    string `json:"login"`
	Password string `json:"password"`
}

func GetToken(c *fiber.Ctx) string {
	token := string(c.Request().Header.Peek("Authorization"))
	if token == "" {
		token = string(c.Request().Header.Cookie(config.AccessTokenCookieName))
	} else {
		strArr := strings.Split(token, " ")
		if len(strArr) == 2 {
			token = strArr[1]
		}
	}
	return token
}

func createToken(userLogin string, userName string) (*TokenDetails, error) {
	td := &TokenDetails{}
	td.AtExpires = time.Now().Add(time.Minute * config.AccessTokenExpiresTime).Unix()
	td.AccessUuid = uuid.New().String()

	td.RtExpires = time.Now().Add(time.Hour * config.RefreshTokenExpiresTime).Unix()
	td.RefreshUuid = uuid.New().String()

	db := database.DBConn
	var groups []models.Group
	result := db.Find(&groups)

	userGroups := []string{}

	if result.RowsAffected > 0 {
		for _, group := range groups {
			gusers := strings.Split(group.Users, ",")
			for _, user := range gusers {
				if user == userLogin {
					userGroups = append(userGroups, group.Name)
				}
			}
		}
	}

	var err error
	//Creating Access Token
	atClaims := jwt.MapClaims{}
	atClaims["authorized"] = true
	atClaims["access_uuid"] = td.AccessUuid
	atClaims["login"] = userLogin
	atClaims["exp"] = td.AtExpires
	atClaims["name"] = userName
	atClaims["groups"] = userGroups

	at := jwt.NewWithClaims(jwt.SigningMethodHS256, atClaims)
	td.AccessToken, err = at.SignedString([]byte(os.Getenv("ACCESS_SECRET")))
	if err != nil {
		return nil, err
	}
	//Creating Refresh Token
	rtClaims := jwt.MapClaims{}
	rtClaims["refresh_uuid"] = td.RefreshUuid
	rtClaims["login"] = userLogin
	rtClaims["exp"] = td.RtExpires
	rtClaims["name"] = userName

	rt := jwt.NewWithClaims(jwt.SigningMethodHS256, rtClaims)
	td.RefreshToken, err = rt.SignedString([]byte(os.Getenv("REFRESH_SECRET")))
	if err != nil {
		return nil, err
	}
	return td, nil
}

func saveRefreshToken(Login string, Devid string, Jti string, RefreshToken string, Exp int) {
	db := database.DBConn
	token := models.Token{}
	token.Login = Login
	token.Devid = Devid
	token.Jti = Jti
	token.RefreshToken = RefreshToken
	token.Exp = Exp
	db.Create(&token)
}

func CheckToken(bearToken string) bool {
	_, err := jwt.Parse(bearToken, func(token *jwt.Token) (interface{}, error) {
		//Make sure that the token method conform to "SigningMethodHMAC"
		if _, ok := token.Method.(*jwt.SigningMethodHMAC); !ok {
			return nil, fmt.Errorf("unexpected signing method: %v", token.Header["alg"])
		}
		return []byte(os.Getenv("ACCESS_SECRET")), nil
	})

	return err == nil
}

func getDevId(c *fiber.Ctx) string {
	devid := string(c.Request().Header.Peek("sessionid")) + string(c.Request().Header.Peek("User-Agent")) + config.SECRET_KEY
	hash := md5.Sum([]byte(devid))
	devid = hex.EncodeToString(hash[:])
	return devid
}

func Login(c *fiber.Ctx) error {
	db := database.DBConn
	auth := AuthLogin{}
	model := models.User{}

	// Parse login data
	if err := c.BodyParser(&auth); err != nil {
		return c.Status(200).JSON(fiber.Map{"errors": []string{"Review your input"}})
	}

	// Calculate the hash of password
	hash := md5.Sum([]byte(auth.Password))
	auth.Password = hex.EncodeToString(hash[:])

	// Search user in table
	result := db.Where("login = ? AND password = ?", auth.Login, auth.Password).First(&model)
	if result.RowsAffected > 0 {
		// Generate Token here
		ts, err := createToken(model.Login, model.Name)
		if err != nil {
			return c.Status(200).JSON(fiber.Map{"errors": []string{"Review your input"}})
		}

		devid := getDevId(c)
		tokens := Tokens{}
		tokens.AccessToken = ts.AccessToken
		tokens.RefreshToken = ts.RefreshToken

		// Save token to database
		saveRefreshToken(auth.Login, devid, ts.RefreshUuid, tokens.RefreshToken, int(ts.RtExpires))

		// Return tokens
		return c.JSON(tokens)
	} else {
		return c.Status(200).JSON(fiber.Map{"errors": []string{"Bad Login or Password"}})
	}
}

func LoginOAuth(c *fiber.Ctx, login string) error {
	db := database.DBConn
	model := models.User{}

	// Search user in table
	result := db.Where("login = ?", login).First(&model)
	if result.RowsAffected > 0 {
		// Generate Token here
		ts, err := createToken(model.Login, model.Name)
		if err != nil {
			return c.Status(200).JSON(fiber.Map{"errors": []string{"Review your input"}})
		}

		devid := getDevId(c)
		tokens := Tokens{}
		tokens.AccessToken = ts.AccessToken
		tokens.RefreshToken = ts.RefreshToken

		// Save token to database
		saveRefreshToken(login, devid, ts.RefreshUuid, tokens.RefreshToken, int(ts.RtExpires))

		// Return tokens
		return c.JSON(tokens)
	} else {
		return c.Status(200).JSON(fiber.Map{"errors": []string{"Bad Login or Password"}})
	}
}

func Refresh(c *fiber.Ctx) error {
	db := database.DBConn
	model := models.Token{}

	tokenString := GetToken(c)

	claims := jwt.MapClaims{}
	_, err := jwt.ParseWithClaims(tokenString, claims, func(token *jwt.Token) (interface{}, error) {
		return []byte(os.Getenv("REFRESH_SECRET")), nil
	})

	if err != nil {
		return c.Status(200).JSON(fiber.Map{"errors": []string{"Unexpected Error"}})
	}

	devid := getDevId(c)
	login := claims["login"]
	jti := claims["refresh_uuid"]
	name := claims["name"]

	// Delete expired refresh tokens
	timestamp := time.Now().Unix()
	db.Where("exp < ?", timestamp).Delete(&model)

	// Search user in table
	result := db.Where("login = ? AND devid = ? AND jti = ? ", login, devid, jti).First(&model)

	if result.RowsAffected > 0 {
		// Generate Token
		ts, err := createToken(login.(string), name.(string))
		if err != nil {
			return c.Status(200).JSON(fiber.Map{"errors": []string{"Unexpected Error"}})
		}

		tokens := Tokens{}
		tokens.AccessToken = ts.AccessToken
		tokens.RefreshToken = ts.RefreshToken

		// Delete current (used once) refresh token
		db.Delete(&model)

		// Save Refresh tokeb to database
		saveRefreshToken(login.(string), devid, ts.RefreshUuid, tokens.RefreshToken, int(ts.RtExpires))

		// Return new tokens
		return c.JSON(tokens)
	} else {
		return c.Status(401).JSON(fiber.Map{})
	}
}

func Logout(c *fiber.Ctx) error {
	// Check Token
	if !CheckToken(GetToken(c)) {
		return c.Status(401).JSON(fiber.Map{})
	}

	db := database.DBConn
	model := models.Token{}

	tokenString := GetToken(c)

	claims := jwt.MapClaims{}
	_, err := jwt.ParseWithClaims(tokenString, claims, func(token *jwt.Token) (interface{}, error) {
		return []byte(os.Getenv("ACCESS_SECRET")), nil
	})

	if err != nil {
		return c.Status(401).JSON(fiber.Map{})
	}

	devid := getDevId(c)
	login := claims["login"]

	db.Where("login = ? AND devid = ?", login, devid).Delete(&model)

	return c.Status(403).JSON(fiber.Map{"msg": []string{"OK"}})
}

func LogoutAll(c *fiber.Ctx) error {
	db := database.DBConn
	model := models.Token{}

	tokenString := GetToken(c)

	claims := jwt.MapClaims{}
	_, err := jwt.ParseWithClaims(tokenString, claims, func(token *jwt.Token) (interface{}, error) {
		return []byte(os.Getenv("ACCESS_SECRET")), nil
	})

	if err != nil {
		return c.Status(401).JSON(fiber.Map{})
	}

	login := claims["login"]

	db.Where("login = ?", login).Delete(&model)

	return c.Status(403).JSON(fiber.Map{"msg": []string{"OK"}})
}
