package main

import (
	"encoding/base64"
	"goapp/access"
	"goapp/auth"
	"goapp/config"
	"goapp/database"
	"goapp/models"
	"goapp/oauth"
	"net/url"
	"strconv"
	"strings"

	"fmt"

	"github.com/gofiber/fiber/v2"
	"github.com/gofiber/fiber/v2/middleware/cors"
	"gorm.io/driver/postgres"
	"gorm.io/gorm"
)

func initDatabase() {
	var err error

	///// Sqlite
	//database.DBConn, err = gorm.Open(sqlite.Open("db.sqlite"), &gorm.Config{})
	/////

	///// Postgres
	dsn := "host=localhost user=postgres password=postgres dbname=test port=5432 sslmode=disable TimeZone=Europe/Kiev"
	database.DBConn, err = gorm.Open(postgres.Open(dsn), &gorm.Config{})
	/////

	if err != nil {
		panic("Failed to connect database")
	}
	fmt.Println("Connection Opened to Database")
	database.DBConn.AutoMigrate(&models.Client{}, &models.Country{}, &models.Group{}, &models.Product{}, &models.Seller{}, &models.User{}, &models.Token{})
	fmt.Println("Database Migrated")

	//Fill the database with test data if the database is empty
	models.CheckAndFill()

}

func setupRoutes(app *fiber.App) {
	app.Static("/", "./public")

	app.Post("/auth/login/", auth.Login)
	app.Post("/auth/refresh/", auth.Refresh)
	app.Get("/auth/logout/", auth.Logout)
	app.Get("/auth/logout-all/", auth.LogoutAll)

	app.Get("/menu", func(c *fiber.Ctx) error {
		// Check Token
		if !auth.CheckToken(auth.GetToken(c)) {
			return c.Status(401).JSON(fiber.Map{})
		}
		// Check Access
		if !access.CheckAccess(c, "menu", auth.GetToken(c)) {
			return c.Status(403).JSON(fiber.Map{"errors": []string{"Access deny"}})
		}
		return models.MenuGets(c)
	})

	app.Get("/:model", func(c *fiber.Ctx) error {
		// Check Token
		if !auth.CheckToken(auth.GetToken(c)) {
			return c.Status(401).JSON(fiber.Map{})
		}
		// Check Access
		if !access.CheckAccess(c, c.Params("model"), auth.GetToken(c)) {
			return c.Status(403).JSON(fiber.Map{"errors": []string{"Access deny"}})
		}
		return models.Gets(c)
	})

	//app.Get("/:model/:id", models.Get)
	app.Get("/:model/:prm", func(c *fiber.Ctx) error {
		// Check Token
		if !auth.CheckToken(auth.GetToken(c)) {
			return c.Status(401).JSON(fiber.Map{})
		}
		// Check Access
		if !access.CheckAccess(c, c.Params("model"), auth.GetToken(c)) {
			return c.Status(403).JSON(fiber.Map{"errors": []string{"Access deny"}})
		}

		prm, err := strconv.Atoi(c.Params("prm"))
		if (err == nil) && (prm > -1) {
			return models.Get(c)
		} else {
			return models.Gets(c)
		}
	})
	app.Post("/:model", func(c *fiber.Ctx) error {
		// Check Token
		if !auth.CheckToken(auth.GetToken(c)) {
			return c.Status(401).JSON(fiber.Map{})
		}
		// Check Access
		if !access.CheckAccess(c, c.Params("model"), auth.GetToken(c)) {
			return c.Status(403).JSON(fiber.Map{"errors": []string{"Access deny"}})
		}
		return models.Create(c)
	})
	app.Put("/:model/:id", func(c *fiber.Ctx) error {
		// Check Token
		if !auth.CheckToken(auth.GetToken(c)) {
			return c.Status(401).JSON(fiber.Map{})
		}
		// Check Access
		if !access.CheckAccess(c, c.Params("model"), auth.GetToken(c)) {
			return c.Status(403).JSON(fiber.Map{"errors": []string{"Access deny"}})
		}
		return models.Update(c)
	})
	app.Delete("/:model/:id", func(c *fiber.Ctx) error {
		// Check Token
		if !auth.CheckToken(auth.GetToken(c)) {
			return c.Status(401).JSON(fiber.Map{})
		}
		// Check Access
		if !access.CheckAccess(c, c.Params("model"), auth.GetToken(c)) {
			return c.Status(403).JSON(fiber.Map{"errors": []string{"Access deny"}})
		}
		return models.Delete(c)
	})

	app.Get("/uploads/:model/:file", func(c *fiber.Ctx) error {
		// Check Token
		if !auth.CheckToken(auth.GetToken(c)) {
			return c.Status(401).JSON(fiber.Map{})
		}
		// Check Access
		if !access.CheckAccess(c, c.Params("model"), auth.GetToken(c)) {
			return c.Status(403).JSON(fiber.Map{"errors": []string{"Access deny"}})
		}
		return models.Uploads(c)
	})

	// OAUTH
	app.Get("/oauth/login/:provider", func(c *fiber.Ctx) error {
		return c.Redirect(oauth.GetAuthURI(c.Params("provider")), 301)
	})

	app.Get("/oauth/:provider/login", func(c *fiber.Ctx) error {
		uri := c.Request().URI().String()
		code := base64.StdEncoding.EncodeToString([]byte(uri))
		return c.Redirect("/frontend/#/oauth/"+c.Params("provider")+"/"+code, 301)
	})

	app.Post("/oauth/login/", func(c *fiber.Ctx) error {
		type OAuthLogin struct {
			Provider string `json:"provider"`
			Code     string `json:"code"`
		}
		var model OAuthLogin
		if err := c.BodyParser(&model); err != nil {
			return c.Status(200).JSON(fiber.Map{"errors": []string{"Unexpected Error"}})
		}

		modelCode, err := base64.StdEncoding.DecodeString(model.Code)
		if err != nil {
			return c.Status(200).JSON(fiber.Map{"errors": []string{"Unexpected Error"}})
		}
		model.Code = string(modelCode)
		model.Code = strings.Replace(model.Code, "http://localhost:5000/", "https://blog.avislab.com/", 3) // WTF? It's lifehack! Read more: https://blog.avislab.com/flask-vue/example9/

		// Get Code
		u, err := url.Parse(model.Code)
		if err != nil {
			return c.Status(200).JSON(fiber.Map{"errors": []string{"Incorrect oauth code"}})
		}
		m, _ := url.ParseQuery(u.RawQuery)
		code := m["code"][0]

		login := oauth.CheckAuth_GetUserID(model.Code, code, model.Provider)
		if login != "" {
			return auth.LoginOAuth(c, login)
		}

		return c.Status(200).JSON(fiber.Map{"errors": []string{"Not authorized"}})
	})
}

func main() {
	config.Load()

	app := fiber.New()

	// Default config
	app.Use(cors.New())

	// Extend your config for customization
	app.Use(cors.New(cors.Config{
		AllowOrigins:  "*",
		AllowHeaders:  "Origin, Content-Type, Accept",
		ExposeHeaders: "*",
	}))

	initDatabase()
	setupRoutes(app)
	access.Init()

	app.Listen(config.APP_LISTEN)

}
