package models

import (
	"encoding/json"
	"fmt"
	"goapp/config"
	"goapp/database"
	"mime/multipart"
	"net/url"
	"reflect"
	"strconv"
	"strings"

	"github.com/gofiber/fiber/v2"
	"github.com/gofiber/fiber/v2/utils"
)

type SearchItem struct {
	Field    string `json:"field"`
	Value    string `json:"value"`
	Operator string `json:"operator"`
}

type Paginator struct {
	Page  int `json:"page"`
	Limit int `json:"limit"`
}

type Prms struct {
	Search    []SearchItem `json:"search"`
	Order     []string     `json:"order"`
	Paginator Paginator    `json:"paginator"`
}

//=========================================
func getModelStruct(modelName string) interface{} {
	switch modelName {
	case "client":
		var value Client
		return &value
	case "countries":
		var value Country
		return &value
	case "groups":
		var value Group
		return &value
	case "product":
		var value Product
		return &value
	case "seller":
		var value Seller
		return &value
	case "users":
		var value User
		return &value
	default:
		return nil
	}
}

func getModelStructArray(modelName string) interface{} {
	switch modelName {
	case "client":
		var value []Client
		return &value
	case "countries":
		var value []Country
		return &value
	case "groups":
		var value []Group
		return &value
	case "product":
		var value []Product
		return &value
	case "seller":
		var value []Seller
		return &value
	case "users":
		var value []User
		return &value
	default:
		return nil
	}
}

func isMultipartForm(c *fiber.Ctx) bool {
	ctype := utils.ToLower(utils.UnsafeString(c.Request().Header.ContentType()))
	return (strings.Index(ctype, "multipart/form-data;") == 0)
}

func multipartSaveFiles(c *fiber.Ctx, form *multipart.Form, modelName string, id reflect.Value) error {
	for fname := range form.File {
		files := form.File[fname]
		for _, file := range files {
			filepath := config.UPLOADS_FOLDER + modelName + "/"
			filename := fmt.Sprintf("%d-%s-%s", id, fname, file.Filename)
			fullfilename := filepath + filename
			// Save the files to disk
			if err := c.SaveFile(file, fullfilename); err != nil {
				return err
			}
		}
	}
	return nil
}

//=========================================

func Gets(c *fiber.Ctx) error {
	modelName := c.Params("model")
	prm := c.Params("prm")
	Prm := Prms{}

	decodedValue, err := url.QueryUnescape(prm)
	if err != nil {
		return c.Status(500).JSON(fiber.Map{"errors": []string{"Invalid params"}})
	} else {
		json.Unmarshal([]byte(decodedValue), &Prm)
	}

	db := database.DBConn
	model := getModelStructArray(modelName)

	// Order
	query := db.Order("")
	count := len(Prm.Order)
	for i := 0; i < count; i++ {
		query = query.Order(Prm.Order[i])
	}

	// Search
	for i := 0; i < len(Prm.Search); i++ {
		field := "CAST(" + Prm.Search[i].Field + " as TEXT)"
		operator := Prm.Search[i].Operator
		value := Prm.Search[i].Value
		if operator == "LIKE" {
			value = "%" + value + "%"
		}
		if i == 0 {
			query = query.Where(field+" "+operator+" ?", "%"+value+"%")
		} else {
			query = query.Or(field+" "+operator+" ?", "%"+value+"%")
		}
	}

	// Paginator
	var total_records int64
	count_query := query.Model(model)
	result := count_query.Find(model)
	total_records = result.RowsAffected

	if Prm.Paginator.Limit > 0 && Prm.Paginator.Page > 0 {
		perpage := Prm.Paginator.Limit
		offset := (Prm.Paginator.Page - 1) * perpage
		query = query.Limit(perpage).Offset(offset)
	}

	query.Find(model)
	c.Append("X-Total-Records", strconv.FormatInt(total_records, 10))

	return c.JSON(model)
}

func Get(c *fiber.Ctx) error {
	modelName := c.Params("model")
	id := c.Params("prm")

	db := database.DBConn
	model := getModelStruct(modelName)
	db.Find(model, id)
	return c.JSON(model)
}

func Create(c *fiber.Ctx) error {
	modelName := c.Params("model")

	db := database.DBConn
	model := getModelStruct(modelName)

	if isMultipartForm(c) {
		if form, err := c.MultipartForm(); err == nil {
			data := form.Value["data"][0]
			json.Unmarshal([]byte(data), model)
			db.Create(model)
			modelId := reflect.ValueOf(model).Elem().FieldByName("ID")
			multipartSaveFiles(c, form, modelName, modelId)
		} else {
			return c.Status(500).JSON(fiber.Map{"errors": []string{"Review your input"}})
		}
	} else {
		if err := c.BodyParser(model); err != nil {
			return c.Status(500).JSON(fiber.Map{"errors": []string{"Review your input"}})
		}
		db.Create(model)
	}
	return c.JSON(model)
}

func Update(c *fiber.Ctx) error {
	modelName := c.Params("model")
	id := c.Params("id")

	db := database.DBConn
	model := getModelStruct(modelName)

	result := db.Find(model, id)
	if result.RowsAffected == 0 {
		return c.Status(404).JSON(fiber.Map{"errors": []string{"Not found"}})
	}

	if isMultipartForm(c) {
		if form, err := c.MultipartForm(); err == nil {
			data := form.Value["data"][0]
			json.Unmarshal([]byte(data), model)
			modelId := reflect.ValueOf(model).Elem().FieldByName("ID")
			multipartSaveFiles(c, form, modelName, modelId)
			db.Save(model)
		} else {
			return c.Status(500).JSON(fiber.Map{"errors": []string{"Review your input"}})
		}
	} else {
		err := c.BodyParser(model)
		if err != nil {
			return c.Status(500).JSON(fiber.Map{"errors": []string{"Review your input"}})
		}
		db.Save(model)
	}
	return c.JSON(model)
}

func Delete(c *fiber.Ctx) error {
	modelName := c.Params("model")
	id := c.Params("id")

	db := database.DBConn
	model := getModelStruct(modelName)
	db.First(model, id)
	if model == nil {
		return c.Status(500).JSON(fiber.Map{"errors": []string{"No " + modelName + " Found with ID"}})
	}
	db.Delete(model)
	return c.JSON(model)
}

func Uploads(c *fiber.Ctx) error {
	modelName := c.Params("model")
	fileName := c.Params("file")
	filepath := config.UPLOADS_FOLDER + modelName + "/" + fileName
	return c.SendFile(filepath)
}

//=========================================

func CheckAndFill() {
	db := database.DBConn

	var user User
	var group Group

	result := db.First(&user, 1)
	if result.RowsAffected == 0 {
		fmt.Println("Database is Empty")

		user.ID = 0
		user.Login = "admin"
		user.Password = "admin"
		user.Name = "Administrator"
		db.Create(&user)

		user.ID = 0
		user.Login = "user"
		user.Password = "123"
		user.Name = "User"
		db.Create(&user)

		fmt.Println("Users 'admin', 'user' created")

		group.ID = 0
		group.Name = "administrators"
		group.Users = "admin,andre,user@gmail.com"
		db.Create(&group)

		group.ID = 0
		group.Name = "operator"
		group.Users = "admin,user,FB User Name"
		db.Create(&group)

		fmt.Println("Groups created")

	} else {
		fmt.Println("Database OK")
	}

}
